import gleam/bit_array
import gleam/dynamic/decode.{type Decoder}
import gleam/int
import gleam/json.{type Json}
import gleam/pair
import gleam/result
import gleam/time/calendar.{
  type Month, April, August, Date, December, February, January, July, June,
  March, May, November, October, September,
}

pub type Date =
  calendar.Date

pub fn parse(value: String) -> Result(Date, Nil) {
  let bytes = bit_array.from_string(value)
  use #(year, bytes) <- result.try(parse_year(from: bytes))
  use bytes <- result.try(accept_byte(from: bytes, value: byte_minus))
  use #(month, bytes) <- result.try(parse_month_calendar(from: bytes))
  use bytes <- result.try(accept_byte(from: bytes, value: byte_minus))
  use #(day, bytes) <- result.try(parse_day_calendar(from: bytes, year:, month:))
  use Nil <- result.try(accept_empty(bytes))
  Ok(Date(year:, month:, day:))
}

pub fn decoder() -> Decoder(Date) {
  use value <- decode.then(decode.string)
  case parse(value) {
    Ok(date) -> decode.success(date)
    Error(Nil) -> decode.failure(Date(1970, calendar.January, 1), "Date")
  }
}

pub fn encode(date: Date) -> Json {
  json.string(
    int.to_string(date.year)
    <> "-"
    <> int.to_string(calendar.month_to_int(date.month))
    <> "-"
    <> int.to_string(date.day),
  )
}

const byte_zero: Int = 0x30

const byte_nine: Int = 0x39

const byte_minus: Int = 0x2D

fn accept_byte(from bytes: BitArray, value value: Int) -> Result(BitArray, Nil) {
  case bytes {
    <<byte, remaining_bytes:bytes>> if byte == value -> Ok(remaining_bytes)
    _ -> Error(Nil)
  }
}

fn accept_empty(from bytes: BitArray) -> Result(Nil, Nil) {
  case bytes {
    <<>> -> Ok(Nil)
    _ -> Error(Nil)
  }
}

fn parse_digits(
  from bytes: BitArray,
  count count: Int,
) -> Result(#(Int, BitArray), Nil) {
  do_parse_digits(from: bytes, count:, acc: 0, k: 0)
}

fn do_parse_digits(
  from bytes: BitArray,
  count count: Int,
  acc acc: Int,
  k k: Int,
) -> Result(#(Int, BitArray), Nil) {
  case bytes {
    _ if k >= count -> Ok(#(acc, bytes))
    <<byte, remaining_bytes:bytes>> if byte_zero <= byte && byte <= byte_nine ->
      do_parse_digits(
        from: remaining_bytes,
        count:,
        acc: acc * 10 + { byte - 0x30 },
        k: k + 1,
      )
    _ -> Error(Nil)
  }
}

fn parse_year(from bytes: BitArray) -> Result(#(Int, BitArray), Nil) {
  parse_digits(from: bytes, count: 4)
}

// slightly modified version of parse_month that returns calendar.Month instead of Int
fn parse_month_calendar(from bytes: BitArray) -> Result(#(Month, BitArray), Nil) {
  use #(month, bytes) <- result.try(parse_digits(from: bytes, count: 2))
  calendar.month_from_int(month) |> result.map(pair.new(_, bytes))
}

// slightly modified version of parse_day that takes calendar.Month instead of Int
fn parse_day_calendar(
  from bytes: BitArray,
  year year: Int,
  month month: Month,
) -> Result(#(Int, BitArray), Nil) {
  use #(day, bytes) <- result.try(parse_digits(from: bytes, count: 2))
  let max_day = case month {
    January | March | May | July | August | October | December -> 31
    April | June | September | November -> 30
    February -> {
      case is_leap_year(year) {
        True -> 29
        False -> 28
      }
    }
  }
  case 1 <= day && day <= max_day {
    True -> Ok(#(day, bytes))
    False -> Error(Nil)
  }
}

fn is_leap_year(year: Int) -> Bool {
  year % 4 == 0 && { year % 100 != 0 || year % 400 == 0 }
}
